/*******************************************************************************
 * VCS client ANE windows native program 
 *******************************************************************************
 * Verson:     0.9
 * Data:       2018.2.10
 * Dscription: Client ANE windows native library for video conference system
 *
 *******************************************************************************/
#define ANE_DBG 0

/*-----------------------------------------------------------------------------*/
/*                          Include Files          	                       */
/*-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#ifdef ANE_WIN_8_10
#include <ShellScalingApi.h>
#endif
#include "FlashRuntimeExtensions.h"

/*-----------------------------------------------------------------------------*/
/*                          External References       	                       */
/*-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------*/
/*                          Constant Definitions                               */
/*-----------------------------------------------------------------------------*/
#define ANE_WINDOW_MAX_NUM         20
#define ANE_NAME_MAX_SIZE          (128*6) /*(bytes, 6bytes/UTF-8)*/
#define ANE_WINDOW_FRAME_WIDTH     8
#define ANE_CHILD_WINDOW_MAX_NUM   20

/*-----------------------------------------------------------------------------*/
/*                          Macro Definitions                                  */
/*-----------------------------------------------------------------------------*/
#if ANE_DBG
#define PRINTD(...) \
    printf( __VA_ARGS__)
#else
#define PRINTD(...) \
    {}
#endif
/*-----------------------------------------------------------------------------*/
/*                          Type & Name Definitions                            */
/*-----------------------------------------------------------------------------*/
#define EXPORT __declspec(dllexport) 
                                            /*---------------------------------*/
                                            /* Window List Element             */
                                            /*---------------------------------*/
typedef struct
{
    HWND           hWnd;                    /* Window handle                   */
    unsigned char  aTitle[ANE_NAME_MAX_SIZE];/* Window Title                   */
} ANE_WINDOW_ELE;

/*-----------------------------------------------------------------------------*/
/*                          Global Functions          	                       */
/*-----------------------------------------------------------------------------*/
EXPORT void vcsExtensionInitializer(
                void **extDataToSet,
                FREContextInitializer *ctxInitializerToSet,
                FREContextFinalizer *ctxFinalizerToSet);
EXPORT void vcsExtensionFinalizer(void *extData);
void vcsContextInitializer(void *extData, const uint8_t *ctxType, 
                           FREContext ctx, uint32_t *numFunctionsToSet,
                           const FRENamedFunction **functionsToSet);
void vcsContextFinalizer(void *extData);
FREObject vcsGetWindowList(FREContext ctx, void* funcData,
                           uint32_t argc, FREObject argv[]);
BOOL CALLBACK vcsCallbackEnumWindow(HWND hWnd, LPARAM lParam);
FREObject vcsGetWindowBitmapData(FREContext ctx, void* funcData,
                                 uint32_t argc, FREObject argv[]);

/*-----------------------------------------------------------------------------*/
/*                          Local Functions          	                       */
/*-----------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------------*/
/*                          Global Varibles                                    */
/*-----------------------------------------------------------------------------*/
FREContext      tVcsExtentionContext;       /* Extension cntext value          */
int             iVcsWindowCount;            /* Window counter                  */
ANE_WINDOW_ELE  aVcsWindowList[ANE_WINDOW_MAX_NUM]; /* Window list             */
unsigned char   aVcsNameBuf[ANE_NAME_MAX_SIZE];     /* Name buffer             */
unsigned char   aVcsTmpBuf[ANE_NAME_MAX_SIZE];      /* Temp. buffer            */
HWND            tVcsMyHwnd;                 /* My window handle                */

                                            /*---------------------------------*/
                                            /* Target Window Control variables */
                                            /*---------------------------------*/
HWND            tVcsTargetHwnd;             /* Target window handle            */
int             iVcsTargetCapturable;       /* Target window type              */
                                            /*     0: Init                     */
                                            /*     1: capturable               */
                                            /*    -1: uncapturable             */

/*-----------------------------------------------------------------------------*/
/*                          Global Constants         	                       */
/*-----------------------------------------------------------------------------*/
                                            /*---------------------------------*/
                                            /* Class Method Function Table     */
                                            /*---------------------------------*/
FRENamedFunction aVcsExtentionMethod[] = 
{
    {                                       /* getWindowList() method          */
        (const uint8_t*)"getWindowList",    /*   name (method name)            */
        NULL,                               /*   functionData                  */
        vcsGetWindowList                    /*   function                      */
    },
    {                                       /* getWindowBitmapData() method    */
        (const uint8_t*)"getWindowBitmapData",/* name (method name)            */
        NULL,                               /*   functionData                  */
        vcsGetWindowBitmapData              /*   function                      */
    }
};

/*-----------------------------------------------------------------------------*
 *     vcsExtensionInitializer()        Extension initializer function
 *-----------------------------------------------------------------------------*
 * Function   Initalize the ANE extension.
 * Entry      extDataToSet       : Pointer to set extension data
 *            ctxInitializerToSet: Pointer to set extension context intializer
 *            ctxFinalizerToSet  : Pointer to set extension context finalizer 
 * Exit       none
 */
EXPORT void vcsExtensionInitializer(
                void **extDataToSet,
                FREContextInitializer *ctxInitializerToSet,
                FREContextFinalizer *ctxFinalizerToSet)
{
    PRINTD("ANE: Init called.\n");

    /*---------------------------------*/
    /* Init ANE extension              */
    /*---------------------------------*/
    *extDataToSet = NULL;                           /* No extension data       */
    *ctxInitializerToSet = vcsContextInitializer;   /* Set context initializer */
    *ctxFinalizerToSet = vcsContextFinalizer;       /* Set context finalizer   */
    tVcsMyHwnd = NULL;                              /* Reset my window handle  */

    return;
}

/*-----------------------------------------------------------------------------*
 *     vcsExtensionFinalizer()        Extension finalizer function
 *-----------------------------------------------------------------------------*
 * Function   Finalize the ANE extension.
 * Entry      extData : Pointer to extension data
 * Exit       none
 */
EXPORT void vcsExtensionFinalizer(void *extData)
{ 
    return;
}

/*-----------------------------------------------------------------------------*
 *     vcsContextInitializer()        Extenson context initializer function
 *-----------------------------------------------------------------------------*
 * Function   Initalize an ANE extension context.
 * Entry      extData          : Pointer to extension data
 *            cxtType          : Context type string
 *            ctx              : Context value
 *            numFunctionsToSet: Pointer to set number of class functions
 *            functionsToSet   : Pointer to set class functions
 * Exit       none
 */
void vcsContextInitializer(void *extData, const uint8_t *ctxType, 
                           FREContext ctx, uint32_t *numFunctionsToSet,
                           const FRENamedFunction **functionsToSet)
{
    /* Save context value */
    tVcsExtentionContext = ctx;

    /* Set method function table */
    *numFunctionsToSet = 2;                 /* (number of functions) */
    *functionsToSet = aVcsExtentionMethod;  /* (table pointer) */

    return;
}

/*-----------------------------------------------------------------------------*
 *     vcsContextFinalizer()        Extension context finalizer function
 *-----------------------------------------------------------------------------*
 * Function   Initalize an ANE extension context.
 * Entry      ctx              : Context value
 * Exit       none
 */
void vcsContextFinalizer(void *extData)
{
    return;
}

/*-----------------------------------------------------------------------------*
 *     vcsGetWindowList()        Extension class method function
 *-----------------------------------------------------------------------------*
 * Function   Get a list of windows including the desktop window 
 *            as the 1st element.
 * Entry      ctx              : Context value
 *            funcData         : Pointer to extension context data
 *            argc             : Number of arguments (= 0)
 *            argv             : Array of arguments (empty array)
 * Exit       return: Array of (window handle, name string)  (if error NULL)
 */
FREObject vcsGetWindowList(FREContext ctx, void* funcData,
                           uint32_t argc, FREObject argv[])
{
    int              i;
    HDC              hdc;
    HWND             hwnd;
    ANE_WINDOW_ELE   *pwin_ele;
    FREObject        obj_array;
    FREObject        obj_uint32;
    FREObject        obj_str;
    FREResult        rc;
#ifdef ANE_WIN_8_10
    int               iwidth;
    int               iwidth1;
    int               iwidth2;
    int               iheight;
    int               iheight1;
    int               iheight2;
    RECT              rect;
    BOOL              bool;
#endif

    ///* Save current foregroung window handle */
    //tVcsMyHwnd = GetForegroundWindow();

#ifdef ANE_WIN_8_10
    if (tVcsMyHwnd == NULL)  while (1) /* (dummy loop to exit by break) */
    {
        /*---------------------------------*/
        /* Get current window size         */
        /*---------------------------------*/
        /* Get current active window handle */
        hwnd = GetActiveWindow();
        if (hwnd == NULL)
        {
            PRINTD("ANE: GetActiveWindow() failed.\n");
            break;
        }
        /* Save my window handle */
        tVcsMyHwnd = hwnd;

        /* Get desktop window handle */
        hwnd = GetDesktopWindow();
        if (hwnd == NULL)
        {
            PRINTD("ANE: 1st GetDesktopWindow() failed.\n");
            break;
        }
        /* Get desktop window rectangle */
        bool = GetWindowRect(hwnd, &rect);
        if (bool == FALSE)
        {
            PRINTD("ANE: 1st GetWindowRect() failed.\n");
            break;
        }
        iwidth1 = rect.right - rect.left;
        iheight1 = rect.bottom - rect.top;
        if ((iwidth1 <= 0) || (iheight1 <= 0))
        {
            PRINTD("ANE:1st  GetWindowRect() size error.\n");
            break;
        }
        //AAAA
        PRINTD("ANE: Old desktop size  W=%d H=%d\n", iwidth1, iheight1);

        /*---------------------------------*/
        /* Change DPI awareness            */
        /*---------------------------------*/
        SetProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE);

        /*---------------------------------*/
        /* Recover my window size          */
        /*---------------------------------*/
        /* Reget desktop window rectangle */
        bool = GetWindowRect(hwnd, &rect);
        if (bool == FALSE)
        {
            PRINTD("ANE: 2nd GetWindowRect() failed.\n");
            break;
        }
        iwidth2 = rect.right - rect.left;
        iheight2 = rect.bottom - rect.top;
        //AAAA
        PRINTD("ANE: New desktop size  W=%d H=%d\n", iwidth2, iheight2);


        if ((iwidth2 <= 0) || (iheight2 <= 0))
        {
            PRINTD("ANE: 2nd GetWindowRect() size error.\n");
            break;
        }
        if ((iwidth1 != iwidth2) || (iheight1 != iheight2))
        {
            /* Get my window rectangle */
            bool = GetWindowRect(tVcsMyHwnd, &rect);
            if (bool == FALSE)
            {
                PRINTD("ANE: 3rd GetWindowRect() failed.\n");
                break;
            }
            iwidth = rect.right - rect.left;
            iheight = rect.bottom - rect.top;
            //AAAA
            PRINTD("ANE: Logical window size  W=%d H=%d\n", iwidth, iheight);

            if ((iwidth <= 0) || (iheight <= 0))
            {
                PRINTD("ANE: 3rd GetWindowRect() size error.\n");
                break;
            }
            /* Change my window size */
            bool = SetWindowPos(tVcsMyHwnd, NULL, 0, 0, 
                                iwidth * iwidth2 / iwidth1, 
                                iheight * iheight2 / iheight1,
                                SWP_NOMOVE | SWP_NOZORDER);
            if (bool == FALSE)
            {
                PRINTD("ANE: SetWindowPos() failed.\n");
                break;;
            }
        }
        break;
    }
#endif

    /*---------------------------------*/
    /* Get desktop window info.        */
    /*---------------------------------*/
    /* Get desktop window handle */
    hwnd = GetDesktopWindow();

    /* Set window list element */
    pwin_ele = aVcsWindowList;
    pwin_ele->hWnd = hwnd;
    memcpy(pwin_ele->aTitle, "Desktop", strlen("Desktop") + 1);
    iVcsWindowCount = 1;

    /*---------------------------------*/
    /* Get other window info.          */
    /*---------------------------------*/
    EnumWindows((WNDENUMPROC)vcsCallbackEnumWindow, (LPARAM)NULL);

    /*---------------------------------*/
    /* Genrate result Array object     */
    /*---------------------------------*/
    /* Create new array object */
    rc = FRENewObject((const uint8_t *)"Array", 0, NULL, &obj_array, NULL);
    if (rc != FRE_OK)
    {
        PRINTD("ANE: Reult Array object creation failed (%x).\n", rc);
        return NULL;
    }
    /* Save window info in Array obj */
    for (i = 0; i < iVcsWindowCount; i++)
    {
        /* Get window list pointer */
        pwin_ele = &aVcsWindowList[i];

        /* Create Uint32 object of window handle */
        rc = FRENewObjectFromUint32((uint32_t)pwin_ele->hWnd, &obj_uint32);
        if (rc != FRE_OK)
        {
            PRINTD("ANE: Window handle (%d) creation failed (%x).\n", i, rc);
            return NULL;
        }
        /* Save window handle in Array obj (even index) */
        rc = FRESetArrayElementAt(obj_array, i * 2, obj_uint32);
        if (rc != FRE_OK)
        {
            PRINTD("ANE: Window handle (%d) saving failed (%x).\n", i, rc);
            return NULL;
        }
        /* Create String object of window title */
        rc = FRENewObjectFromUTF8(strlen(pwin_ele->aTitle) + 1, 
                                  (const uint8_t *)pwin_ele->aTitle, &obj_str);
        if (rc != FRE_OK)
        {
            PRINTD("ANE: Title String (%d) creation failed (%x).\n", i, rc);
            return NULL;
        }
        /* Save title string in Array obj (odd index) */
        rc = FRESetArrayElementAt(obj_array, i * 2 + 1, obj_str);
        if (rc != FRE_OK)
        {
            PRINTD("ANE: Window title (%d) saving failed (%x).\n", i, rc);
            return NULL;
        }
    }
    /* Set length of Array object */
    rc = FRESetArrayLength(obj_array, i * 2);
    if (rc != FRE_OK)
    {
        PRINTD("ANE: Array length saving failed (%x).\n", rc);
        return NULL;
    }
    /* Return successfully */
    return obj_array;
}

/*-----------------------------------------------------------------------------*
 *     vcsCallbackEnumWindow()        Extension function
 *-----------------------------------------------------------------------------*
 * Function   Callback function of EnumWindows() to get a list of windows.
 * Entry      hWnd   : Window handle
 *            lParam : Parameter pointer
 * Exit       return  true:  suceeded
 *                    false: failed
 */
BOOL CALLBACK vcsCallbackEnumWindow(HWND hWnd, LPARAM lParam)
{
    int              i;
    int              ilen;
    int              iret;
    ANE_WINDOW_ELE   *pwin_ele;
    BOOL             bool;
    WINDOWPLACEMENT  twin_placement;
    unsigned char    aclass_name[32];
    RECT             rect;

    /* Check if window count over */
    if (iVcsWindowCount >= ANE_WINDOW_MAX_NUM)
    {
        return FALSE;
    }
    /* Get window list pointer */
    pwin_ele = &aVcsWindowList[iVcsWindowCount];

    /*---------------------------------*/
    /* Check window state              */
    /*---------------------------------*/
    if (!IsWindowVisible(hWnd))
    {
        /* Return if not visible */
        return TRUE;
    }
    if (IsIconic(hWnd))
    {
        /* Return if in icon state */
        return TRUE;
    }
    bool = GetWindowRect(hWnd, &rect);
    if ( (bool == FALSE) ||
         (rect.right - rect.left <= 0) ||
         (rect.bottom - rect.top <= 0) )
    {
        /* Return if size not detected */
        return TRUE;
    }
    if (GetParent(hWnd) != NULL)
    {
        /* Return if parent window exists */
        return TRUE;
    }

    //if (GetWindow(hWnd, GW_OWNER) != NULL) 
    //{
    //    /* Return if owner window exists */
    //    return TRUE;
    //}

    /*---------------------------------*/
    /* Get class name                  */
    /*---------------------------------*/
    ilen = GetClassName(hWnd, aclass_name, sizeof(aclass_name));
    if (ilen == sizeof(aclass_name))
    {
        /* If buffer full, set EOS */
        aclass_name[sizeof(aclass_name) - 1] = '\0';
    }
    if (lstrcmp((LPTSTR)aclass_name, TEXT("Progman")) == 0)
    {
        /* Return if Program Manager's window */
        return TRUE;
    }
    /*---------------------------------*/
    /* Get window title                */
    /*---------------------------------*/
    ilen = GetWindowText(hWnd, (LPTSTR)aVcsNameBuf, ANE_NAME_MAX_SIZE);
    if (ilen == 0)
    {
        /* Return if no title */
        return TRUE;
    }
    if (ilen ==  ANE_NAME_MAX_SIZE)
    {
        /* If buffer full, set EOS */
        ilen--;
        aVcsNameBuf[ilen] = '\0';
    }
    /* Convert char code to wide char (SJIS to UTF-16) */
    iret = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED + MB_ERR_INVALID_CHARS,
                               (LPCSTR)aVcsNameBuf, -1,
                               (LPWSTR)aVcsTmpBuf, 
                               ANE_NAME_MAX_SIZE / 4 - 1);
    /* If not converted, use input string */
    if (iret == 0)
    {
        PRINTD("ANE: Window name conversion failed Title=%.64s\n", 
                aVcsNameBuf);
        memcpy(pwin_ele->aTitle, aVcsNameBuf, ilen + 1); /* (+1: EOS) */
    }
    /* If converted, reconvert wide char to UTF-8 (UTF-16 to UTF-8) */
    else
    {
        ilen = WideCharToMultiByte(CP_UTF8, 0,
                                   (LPCWSTR)aVcsTmpBuf, iret,
                                   (LPSTR)pwin_ele->aTitle, 
                                   ANE_NAME_MAX_SIZE, NULL, NULL);
        if (ilen == 0)
        {
            /* Return if no title */
            return TRUE;
        }
        if (ilen ==  ANE_NAME_MAX_SIZE)
        {
            /* If buffer full, set EOS */
            ilen--;
            pwin_ele->aTitle[ilen] = '\0';
        }
    }
    /*---------------------------------*/
    /* Check if duplicted title        */
    /*---------------------------------*/
    for (i = 1; i < iVcsWindowCount; i++) /* (i=0: desktop) */
    {
        if (strcmp((const char *)&aVcsWindowList[i].aTitle,
                   (const char *)&pwin_ele->aTitle) == 0)  break;
    }
    if (i < iVcsWindowCount)
    {
        /* Return if duplicated title */
        return TRUE;
    }
    /*---------------------------------*/
    /* Save detected window in table   */
    /*---------------------------------*/
    PRINTD("ANE: Window detected %d hWnd=%d Title=%.64s\n", 
           iVcsWindowCount, (int)hWnd, pwin_ele->aTitle);

    /* Save window handle */
    pwin_ele->hWnd = hWnd;

    /* Update window counter */
    iVcsWindowCount++;

    return TRUE;
}

/*-----------------------------------------------------------------------------*
 *     vcsGetWindowBitmapData()        Extension class method function
 *-----------------------------------------------------------------------------*
 * Function   Get BitmapData of a designated window.
 * Entry      ctx              : Context value
 *            funcData         : Pointer to extension context data
 *            argc             : Number of arguments (= 1)
 *            argv             : argv[0]: Handle of target window
 * Exit       return: BitmapData object (if error NULL)
 */
FREObject vcsGetWindowBitmapData(FREContext ctx, void* funcData,
                                 uint32_t argc, FREObject argv[])
{
    int               i;
    int               ilen;
    int               inew_hwnd_flag;
    int               iwidth;
    int               iheight;
    int               isrc_x;
    int               isrc_y;
    unsigned int      *pdot;
    HWND              hwnd;
    //HWND            hwnd_fg;
    RECT              rect;
    BOOL              bool;
    HDC               hdc_win;
    HDC               hdc;
    BITMAPINFO        bmi;
    void              *pvbits;
    HBITMAP           hbmp;
    HBITMAP           hbmp2;
    HBITMAP           hbmp_save;
    HBITMAP           hbmp_save2;
    BITMAP            bmp;
    FREResult         rc;
    FREObject         obj_bitmapdata;
    FREObject         argv_obj[4];
    FREBitmapData2    tbitmapdata2;
#if 0
    unsigned char     *psrc;
    unsigned char     *pdst;
#endif

    /*---------------------------------*/
    /* Get taget window                */
    /*---------------------------------*/
    /* Get target window handle */
    FREGetObjectAsUint32(argv[0], (uint32_t *)&hwnd);

    /* Check if new handle */
    if (hwnd != tVcsTargetHwnd)
    {
        /* Set new window handle flag */
        inew_hwnd_flag = 1;

        /* Save window handle */
        tVcsTargetHwnd = hwnd;

        /* Reset window type */
        iVcsTargetCapturable = 0;
    }
    else
    {
        /* Clear new window handle flag */
        inew_hwnd_flag = 0;
    }

    /* Get base window rectangle */
    bool = GetWindowRect(hwnd, &rect);
    if (bool == FALSE)
    {
        PRINTD("ANE: GetWindowRect(%d) failed.\n",(int)hwnd);
        return NULL;
    }
    iwidth = rect.right - rect.left;
    iheight = rect.bottom - rect.top;
    isrc_x = rect.left;
    isrc_y = rect.top;
    //PRINTD("ANE-DBG: Window rect  W=%d H=%d\n", iwidth, iheight);
    /* If taget window not caputurable, resize it for BitBlt */
    if (iVcsTargetCapturable == -1)
    {
        iwidth -= ANE_WINDOW_FRAME_WIDTH * 2;
        iheight -= ANE_WINDOW_FRAME_WIDTH * 2;
        isrc_x += ANE_WINDOW_FRAME_WIDTH;
        isrc_y += ANE_WINDOW_FRAME_WIDTH;
    }
    //PRINTD("ANE-DBG: Base window  left=%d top=%d right=%d bottom=%d\n",
    //       rect.left, rect.top, rect.right, rect.bottom);

    if ((iwidth <= 0) || (iheight <= 0))
    {
        PRINTD("ANE: getWindowRect(%d) size error.\n",(int)hwnd);
        return NULL;
    }
    /*---------------------------------*/
    /* Get window bitmap data          */
    /*---------------------------------*/
    /* Create device context for buffer */
    hdc = CreateCompatibleDC(NULL);
    if (hdc == NULL)
    {
        PRINTD("ANE: CreateCompatibleDC(%d) failed.\n", (int)hwnd);
        return NULL;
    }
    /* Create DIB (Device independent Bitmap) */
    memset(&bmi, 0, sizeof(BITMAPINFO));
    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = iwidth;
    bmi.bmiHeader.biHeight = iheight;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 32; /* (32bit color) */
    hbmp = CreateDIBSection(NULL, (LPBITMAPINFO)&bmi, DIB_RGB_COLORS, 
                            &pvbits, NULL, 0);
    if (hbmp == NULL)
    {
        PRINTD("ANE: CreateDIBSection failed.\n");

        /* Release resource then return */
        DeleteDC(hdc);
        return NULL;
    }
    /* Select DIB (current bitmap saved) */
    hbmp_save = (HBITMAP)SelectObject(hdc, (HGDIOBJ)hbmp);
    if (hbmp_save == NULL)
    {
        PRINTD("ANE: SelectObject failed.\n");

        /* Release resources then return */
        DeleteObject(hbmp);
        DeleteDC(hdc);
        return NULL;
    }
    /* Copy window bitmap data to DIB */
    if (hwnd == aVcsWindowList[0].hWnd)
    {
        /*---------------------------------*/
        /* Case: Desktop                   */
        /*---------------------------------*/
        /* Get window device context for BitBlt() */
        hdc_win = GetWindowDC(hwnd);
        if (hdc_win == NULL)
        {
            PRINTD("ANE: getWindowDc(%d) failed.\n", (int)hwnd);

            /* Release resources then return */
            SelectObject(hdc, hbmp_save);
            DeleteObject(hbmp);
            DeleteDC(hdc);
            return NULL;
        }
        /* Copy bitmap of desktop by BitBlt */
        bool = BitBlt(hdc, 0, 0, iwidth, iheight, hdc_win, 0, 0, SRCCOPY);
    }
    else
    {
        /*---------------------------------*/
        /* Case: Application window        */
        /*---------------------------------*/
        bool = TRUE;
        while (1) /* (dummy loop to exit by break) */
        {
            /*---------------------------------*/
            /* Copy bitmap by PrintWindow      */
            /*---------------------------------*/
            /* Check if not uncapturable */
            if (iVcsTargetCapturable != -1)
            {
                /* Copy bitmap by PrintWindow */
                bool = PrintWindow(hwnd, hdc, 0);

                /* If new window handle designated, re-copy bitmap */
                if (inew_hwnd_flag != 0)
                {
                    bool = PrintWindow(hwnd, hdc, 0);
                }
                if (bool == FALSE)  break; /* (exit) */
            }
            /* If selected and caputurable window, exit */
            if (iVcsTargetCapturable == 1)  break; /* (exit) */

            /*---------------------------------*/
            /* Check if result is black image  */
            /*---------------------------------*/
            /* Check if not uncapturable window */
            if (iVcsTargetCapturable != -1)
            {
                /* Get window bitmap data info. */
                ilen = GetObject((HGDIOBJ)hbmp, sizeof(BITMAP), (LPVOID)&bmp);
                if (ilen == 0)
                {
                    bool = FALSE;
                    break;  /* (exit) */
                }
                /* Check if black window */
                pdot = (unsigned int *)bmp.bmBits;
                for (i = 0; i < (iwidth * iheight); i++)
                {
                    if (*pdot++ != 0x000000)  break;
                }
                /* If not black window, exit */
                if (i < (iwidth * iheight))
                {
                    //PRINTD("ANE_DBG: Non-black found %d (%d)\n", (int)hwnd, i);

                    /* Set window in capturable */
                    iVcsTargetCapturable = 1;
                    break;  /* (exit) */
                }
                /* Set window in uncapturable */
                iVcsTargetCapturable = -1;

                /* Resize bitmap for BitBlt */
                iwidth -= ANE_WINDOW_FRAME_WIDTH * 2;
                iheight -= ANE_WINDOW_FRAME_WIDTH * 2;
                isrc_x += ANE_WINDOW_FRAME_WIDTH;
                isrc_y += ANE_WINDOW_FRAME_WIDTH;

                /* Re-create DIB of modified size */
                memset(&bmi, 0, sizeof(BITMAPINFO));
                bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
                bmi.bmiHeader.biWidth = iwidth;
                bmi.bmiHeader.biHeight = iheight;
                bmi.bmiHeader.biPlanes = 1;
                bmi.bmiHeader.biBitCount = 32; /* (32bit color) */
                hbmp2 = CreateDIBSection(NULL, (LPBITMAPINFO)&bmi, 
                                         DIB_RGB_COLORS, &pvbits, NULL, 0);
                if (hbmp2 == NULL)
                {
                    PRINTD("ANE: CreateDIBSection failed (2).\n");
                    bool = FALSE;
                    break; /* (exit) */
                }
                /* Re-select DIB (current bitmap saved) */
                hbmp_save2 = (HBITMAP)SelectObject(hdc, (HGDIOBJ)hbmp2);
                if (hbmp_save2 == NULL)
                {
                    PRINTD("ANE: SelectObject failed (2).\n");
                    DeleteObject(hbmp2);
                    bool = FALSE;
                    break; /* (exit) */
                }
                /* Release previous bitmap resource */
                DeleteObject(hbmp_save2);
                hbmp = hbmp2;
            }
            /*---------------------------------*/
            /* Copy from desktop by BitBlt     */
            /*---------------------------------*/
            ///* Get current forground window handle */
            //hwnd_fg = GetForegroundWindow();
            ///* Check if my window not in foreground */
            //if (hwnd_fg != tVcsMyHwnd)
            //{
                /* Raise target window to top of screen */
                //SetForegroundWindow(hwnd);
                ////SetWindowPos(hwnd, HWND_TOP, 0, 0, 0, 0, 
                ////             (SWP_NOMOVE | SWP_NOSIZE));
            //}
            /* Get window device context for BitBlt() */
            hdc_win = GetWindowDC(hwnd);
            if (hdc_win == NULL)
            {
                PRINTD("ANE: getWindowDc (%d) failed (2).\n", (int)hwnd);
                bool = FALSE;
                break; /* (exit) */
            }

            PRINTD("ANE-DBG: Bitmap copyed from desktop.\n");

            /* Get window device context of desktop for BitBlt() */
            hdc_win = GetWindowDC(aVcsWindowList[0].hWnd);
            if (hdc_win == NULL)
            {
                PRINTD("ANE: getWindowDc(%d) failed (2).\n", (int)hwnd);
                bool = FALSE;
                break; /* (exit) */
            }
            /* Copy bitmap from desktop by BitBlt */
            bool = BitBlt(hdc, 0, 0, iwidth, iheight, hdc_win, 
                          isrc_x, isrc_y, SRCCOPY);
            break; /* (exit) */
        }
    }
    if (bool == FALSE)
    {
        PRINTD("ANE: Bitmap capture failed.\n");

        /* Release resources then return */
        SelectObject(hdc, hbmp_save);
        DeleteObject(hbmp);
        DeleteDC(hdc);
        return NULL;
    }
    /*---------------------------------*/
    /* Get window bitmap data info.    */
    /*---------------------------------*/
    ilen = GetObject((HGDIOBJ)hbmp, sizeof(BITMAP), (LPVOID)&bmp);
    if (ilen == 0)
    {
        PRINTD("ANE: GetObject failed.\n");

        /* Release resources then return */
        SelectObject(hdc, hbmp_save);
        DeleteObject(hbmp);
        DeleteDC(hdc);
        return NULL;
    }
    /*---------------------------------*/
    /* Genrate result BitmapData object*/
    /*---------------------------------*/
    obj_bitmapdata = NULL;
    while (1) /* (dummy loop to exit by break) */
    {
        /* Create new BitmapData object */
        rc = FRENewObjectFromInt32(iwidth, &argv_obj[0]);   /* (width) */
        if (rc != FRE_OK)  break; 
        rc = FRENewObjectFromInt32(iheight, &argv_obj[1]); /* (height) */
        if (rc != FRE_OK)  break; 
        rc = FRENewObjectFromBool(0, &argv_obj[2]); /* (transparent:false) */
        if (rc != FRE_OK)  break; 
        rc = FRENewObjectFromUint32(0, &argv_obj[3]);     /* (fillColor:0) */
        if (rc != FRE_OK)  break; 
        rc = FRENewObject((const uint8_t *)"flash.display.BitmapData", 
                          4, argv_obj, &obj_bitmapdata, NULL);
        if (rc != FRE_OK)  break;

        /* Get pointer to access BitmapData object */
        rc = FREAcquireBitmapData2(obj_bitmapdata, &tbitmapdata2);
        if (rc != FRE_OK)  break;
#if 1
        /* Copy window 32bit bitmap data to BitmapData object */
        memcpy((unsigned char *)tbitmapdata2.bits32, 
               (unsigned char *)bmp.bmBits,
               iwidth * iheight * 4);
#else
        /* Copy window 32bit bitmap data to BitmapData object(from bottom line)*/
        ilen = iwidth * 4; /* (line byte length) */
        pdst = (unsigned char *)tbitmapdata2.bits32;
        psrc = (unsigned char *)bmp.bmBits + ilen * (iheight - 1);
        for (i = iheight - 1; i >= 0; i--)
        {
            memcpy(pdst, psrc, ilen);
            pdst += ilen;
            psrc -= ilen;
        }
#endif

#if 1
        /* Notice modified area of BitmapData object (to runtime) */
        rc = FREInvalidateBitmapDataRect(obj_bitmapdata, 0, 0, iwidth, iheight);
        //if (rc != FRE_OK)  break;
#endif
        break;
    }
    /* Release resources */
    SelectObject(hdc, hbmp_save);
    DeleteObject(hbmp);
    DeleteDC(hdc);
    if (obj_bitmapdata != NULL) 
    {
        FREReleaseBitmapData(obj_bitmapdata);
    }
    /* Check error */
    if (rc != FRE_OK)
    { 
        PRINTD("ANE: Bitmap data creation failed.\n");
        return NULL;
    }
    /* Return successfully */
    return obj_bitmapdata;
}

/*-----------------------------------------------------------------------------*/
/*                          End of Program Codes      	                       */
/*-----------------------------------------------------------------------------*/
#if 0

FREReleaseBitmapData()
FREInvalidateBitmapDataRect()


FREResult FREAcquireBitmapData2 (
            FREObject        object,
            FREBitmapData2* descriptorToSet
);

typedef struct tagBITMAP {  /* bm */
    int bmType;  
    int bmWidth;  
    int bmHeight;  
    int bmWidthBytes;  
    BYTE bmPlanes;  
    BYTE bmBitsPixel;  
    LPVOID bmBits;  
} BITMAP;  

FREResult FRENewObjectFromUTF8(
            uint32_t        length,
            const uint8_t* value,
            FREObject*      object
);

FREResult FRESetArrayElementAt (
            FREObject    arrayOrVector,
            uint32_t     index,
            FREObject    value
);

FREResult FRESetArrayLength (
            FREObject    arrayOrVector,
            uint32_t     length
);


FREResult FRENewObject(
            const uint8_t* className,
            uint32_t       argc,
            FREObject      argv[],
            FREObject*     object,
            FREObject*     thrownException
);

enum FREResult {
    FRE_OK                  = 0,
    FRE_NO_SUCH_NAME        = 1,
    FRE_INVALID_OBJECT      = 2,
    FRE_TYPE_MISMATCH       = 3,
    FRE_ACTIONSCRIPT_ERROR  = 4,
    FRE_INVALID_ARGUMENT    = 5,
    FRE_READ_ONLY           = 6,
    FRE_WRONG_THREAD        = 7,
    FRE_ILLEGAL_STATE       = 8,
    FRE_INSUFFICIENT_MEMORY = 9,
    FREResult_ENUMPADDING   = 0xffff
};

enum FREObjectType {
    FRE_TYPE_OBJECT           = 0,
    FRE_TYPE_NUMBER           = 1,
    FRE_TYPE_STRING           = 2,
    FRE_TYPE_BYTEARRAY        = 3,
    FRE_TYPE_ARRAY            = 4,
    FRE_TYPE_VECTOR           = 5,
    FRE_TYPE_BITMAPDATA       = 6,
    FRE_TYPE_BOOLEAN          = 7,
    FRE_TYPE_NULL             = 8,
    FREObjectType_ENUMPADDING = 0xfffff
};
#endif

/*-----------------------------------------------------------------------------*/
/*                          End of Module          	                       */
/*-----------------------------------------------------------------------------*/
